package org.netkernel.client.http.endpoint;

import org.netkernel.client.http.representation.HttpCacheRepresentation;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;

import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

public class SimpleInMemoryCache extends StandardAccessorImpl
{	private Map<Key, CacheEntry> mMap;
	
	private int mReapCount;
	private static int mMaxReapCount=25;
	private static double mMaxSize=8*1024*1024*128;  // Max size in bytes of the aggregate ContentLength of all cached items - 128MB 
	private double mCacheSize;
	
	private int mLastThreshold=0;
	private int mCullPercent=50;

	public SimpleInMemoryCache()
	{	mMap=new ConcurrentHashMap<Key, CacheEntry>(64);		
	}

	@Override
	public void onSink(INKFRequestContext aContext) throws Exception
	{	String url=aContext.getThisRequest().getArgumentValue("url");
		HttpCacheRepresentation rep=(HttpCacheRepresentation)aContext.sourcePrimary(HttpCacheRepresentation.class);
		CacheEntry e=new CacheEntry(rep);
		mMap.put(new Key(url, e), e);
		mCacheSize+=rep.getContentLength();
		mReapCount++;
		if(mReapCount==mMaxReapCount || mCacheSize>mMaxSize)
		{	reap();
		}
	}

	@Override
	public void onSource(INKFRequestContext aContext) throws Exception
	{	Object result=null;
		String mimetype=null;
		String url=aContext.getThisRequest().getArgumentValue("url");
		CacheEntry e=mMap.get(new Key(url));
		if(e!=null)
		{	result=e.getRepresentation();
		}
		INKFResponse resp=aContext.createResponseFrom(result);
		resp.setMimeType(mimetype);
		resp.setNoCache();
		/*
		if(result==null)
		{	resp.setNoCache();			
		}
		*/
	}
	
	private void reap()
	{	mReapCount=0;
		cull();
	}
	
	
	private class CacheEntry
	{	long mLastTouch;
		HttpCacheRepresentation mRep;
		int mHits=1;
		private int mStaticIndex;
	
		public CacheEntry(HttpCacheRepresentation aRep)
		{	mRep=aRep;
			increment(System.currentTimeMillis());
		}
		
		public HttpCacheRepresentation getRepresentation()
		{	increment(System.currentTimeMillis());
			return mRep;
		}
		
		public int getCacheIndex()
		{	return mStaticIndex;
		}
		
		public int update(long aNow)
		{
			long index = ((aNow - mLastTouch)<<4) * (mRep.getContentLength() >> 7 );
			mStaticIndex = (index >= Integer.MAX_VALUE) ? (Integer.MAX_VALUE - 1) : (int) index;
			return mStaticIndex; 
		}
		
		public void increment(long aNow)
		{	mHits++;
			mLastTouch=aNow;
		}
	}
	
	private class Key
	{
		CacheEntry mEntry;
		String mURL;
		
		public Key(String aURL)
		{	this(aURL, null);			
		}
		
		public Key(String aURL, CacheEntry aEntry)
		{	mEntry=aEntry;
			mURL=aURL;
		}
		
		public CacheEntry getEntry()
		{	return mEntry;			
		}
		
		public String getURL()
		{	return mURL;			
		}
		
		public boolean equals(Object aOther)
		{	return (aOther instanceof Key) && mURL.equals(((Key)aOther).getURL() );	
		}
		
		public int hashCode()
		{	return mURL.hashCode();			
		}
		
	}
	
	private int cull()
	{
		int size=mMap.size();
		int removeTarget=size*mCullPercent/100;
		
		long now=System.currentTimeMillis();

		int c1=0,c2=0,c3=0;
		
		int t1=mLastThreshold/2;
		int t2=mLastThreshold;
		int t3=mLastThreshold+mLastThreshold/2;
		
		for (Key key : mMap.keySet() )
		{	
			CacheEntry item=key.getEntry();
			int idx= item.update(now);
			if (idx>t1)
			{	c1++;
				if (idx>t2)
				{	c2++;
					if (idx>t3)
					{	c3++;
					}
				}
			}
		}
		//mLogger.logRaw(ILogger.INFO,this,0+" C("+c1+" "+c2+" "+c3+")");
		//mLogger.logRaw(ILogger.INFO,this,0+" T("+t1+" "+t2+" "+t3+")");
			
		int n=0;
		while(n<3)
		{	
			if (c1<removeTarget) // too few removed
			{	mLastThreshold=mLastThreshold/2;
			}
			else if (c3>removeTarget) // too many removed
			{	mLastThreshold=mLastThreshold*2;
			}
			else
			{	break;
			}
			
			n++;
			
			c1=0;
			c2=0;
			c3=0;
			
			t1=mLastThreshold/2;
			t2=mLastThreshold;
			t3=mLastThreshold+mLastThreshold/2;
			
			for (Key key : mMap.keySet() )
			{	int idx=key.getEntry().getCacheIndex();
				if (idx>t1)
				{	c1++;
					if (idx>t2)
					{	c2++;
						if (idx>t3)
						{	c3++;
						}
					}
				}
			}
		}
		
		long diff;
		if (mLastThreshold==0)
		{	mLastThreshold=1;
		}
		
		if (c2>removeTarget) // threshold needs to get larger, towards t3
		{	diff=(long)mLastThreshold*((long)c2-(long)removeTarget)/(4L*(long)(c2-c3)+1L);
		}
		else
		{	diff=-(long)mLastThreshold*((long)removeTarget-(long)c2)/(2L*(long)(c1-c2)+1L);
		}
		if (diff>mLastThreshold)
		{	diff=mLastThreshold;
		}
		else if (diff<-mLastThreshold)
		{	diff=-mLastThreshold/2;
		}
		mLastThreshold+=(int)diff;
				
		// remove items over threshold
		int removed=0;
		for (Iterator<Key> i=mMap.keySet().iterator(); i.hasNext(); )
		{	Key key=i.next();		
			CacheEntry item=key.getEntry();
			if (item.getCacheIndex()>mLastThreshold)
			{	i.remove();
				mCacheSize =- item.getRepresentation().getContentLength();
				removed++;
			}
		}
	
		long duration=System.currentTimeMillis()-now;
		int removedPercent=removed*100/(size+1);
		
		//mStatus=mLogger.format("MSG_CACHE_STATUS",duration,mGCHeapAvailable,size,mLastThreshold,removedPercent);
		//mLogger.logRaw(ILogger.CACHE,this,mStatus);
		
		System.gc();
		return removed;
	}
}
